
import { HttpClientModule, HttpClient } from '@angular/common/http';
import { TranslateModule, TranslateLoader } from '@ngx-translate/core';
import { TranslateHttpLoader } from '@ngx-translate/http-loader';
import { GoogleMaps } from '@ionic-native/google-maps';
import { ComponentsModule } from '../components/components.module';
import { Connectivity } from '../providers/connectivity/connectivity';
import { GoogleMapsService } from '../providers/google-maps/google-maps';
import { Network } from '@ionic-native/network';
import { LocationsService } from '../providers/locations-service/locations-service';
import { AuthService } from '../providers/auth-service/auth-service';
import { AlertService } from '../providers/alert-service/alert-service';
import { PlacesService } from '../providers/places-service/places-service';
import { Geolocation } from '@ionic-native/geolocation';
import { Camera } from '@ionic-native/camera';
import { CameraService } from '../providers/camera-service/camera-service';
import { Ionic2RatingModule } from 'ionic2-rating';
import { LazyLoadImageModule } from 'ng-lazyload-image';
import * as ionicGalleryModal from 'ionic-gallery-modal';
import { HAMMER_GESTURE_CONFIG } from '@angular/platform-browser';
import { GooglePlus } from '@ionic-native/google-plus';
import { UserServiceProvider } from '../providers/user-service/user-service';
import { MomentModule } from 'angular2-moment';

export function createTranslateLoader(http: HttpClient) {
    return new TranslateHttpLoader(http, './assets/i18n/', '.json');
}



export const MODULES = [
    HttpClientModule,
    TranslateModule.forRoot({
        loader: {
            provide: TranslateLoader,
            useFactory: (createTranslateLoader),
            deps: [HttpClient]
        }
    }),
    ComponentsModule,
    Ionic2RatingModule,
    LazyLoadImageModule,
    ionicGalleryModal.GalleryModalModule,
    MomentModule
]

export const PROVIDERS = [
    GoogleMaps, 
    Connectivity, 
    GoogleMapsService,
    Network,
    LocationsService,
    AuthService,
    AlertService,
    PlacesService,
    Geolocation,
    Camera,
    CameraService,
    {
        provide: HAMMER_GESTURE_CONFIG,
        useClass: ionicGalleryModal.GalleryModalHammerConfig,
    },
    GooglePlus,
    UserServiceProvider
]

