import { LatLng, BaseClass, GoogleMapsEvent, Marker, GoogleMap, Circle } from "@ionic-native/google-maps";
import { PlaceCategory, PlaceDetails, Activity, PlaceComment, PlaceInterface, UserPlace, EntryRegister, PlaceRating } from "../../providers/utils/constants";
import { User } from "@firebase/auth-types";
import { Observable } from "@firebase/util";
import { nullSafeIsEquivalent } from "@angular/compiler/src/output/output_ast";


export class Place extends BaseClass {
    private _uid: string;
    private _name: string;
    private _description: string;
    private _location: LatLng;
    private _place_category: PlaceCategory;
    private _place_details: PlaceDetails;
    private _activities: Activity[];
    private _ratings?: PlaceRating[];
    private _comments?: PlaceComment[];
    private _images?: string[];
    private _imagesUrl?: string[];
    private _marker?: Marker;
    private _user: UserPlace;
    private _created_at: Date;
    private _updated_at: Date;
    private _checkin_registers: EntryRegister[];


    constructor() {
        super();

        this._place_category = {
            lang_key: null,
            uid: null
        };

        this._place_details = {
            free: null,
            water: null,
            wc: null
        };

        this._user = {
            uid: null,
            display_name: null,
            photo_profile: null,
            rating: null
        };

    }

    setData(place: PlaceInterface) {
        this._uid = place.uid;
        this._name = place.name;
        this._description = place.description;
        this._location = place.location;
        this._place_category = place.place_category;
        this._place_details = place.place_details;
        this._activities = place.activities ? place.activities : [];
        this._ratings = place.ratings;
        this._comments = place.comments;
        this._images = place.images ? place.images : [];
        this._imagesUrl = place.imagesUrl ? place.imagesUrl : [];
        this._marker = place.marker;
        this._user = place.user;
        this._created_at = place.created_at;
        this._updated_at = place.updated_at;
        this._checkin_registers = place.checkin_registers;
    }


    setMarker(marker: Marker, map: GoogleMap) {
        this._marker = marker;
        marker.addEventListener(GoogleMapsEvent.MARKER_CLICK).subscribe((value) => {
        });
    }


    getRating(): number {
        let sum = 0;
        for(let i of this.ratings){
            sum = sum + i.rating;
        }
        return sum / this.ratings.length;
    }


    /**
     * Getter uid
     * @return {string}
     */
	public get uid(): string {
		return this._uid;
	}

    /**
     * Getter name
     * @return {string}
     */
	public get name(): string {
		return this._name;
	}

    /**
     * Getter description
     * @return {string}
     */
	public get description(): string {
		return this._description;
	}

    /**
     * Getter place_category
     * @return {PlaceCategory}
     */
	public get place_category(): PlaceCategory {
		return this._place_category;
	}

    /**
     * Getter place_details
     * @return {PlaceDetails}
     */
	public get place_details(): PlaceDetails {
		return this._place_details;
	}

    /**
     * Getter activities
     * @return {Activity[]}
     */
	public get activities(): Activity[] {
		return this._activities;
	}

    /**
     * Setter uid
     * @param {string} value
     */
	public set uid(value: string) {
		this._uid = value;
	}

    /**
     * Setter name
     * @param {string} value
     */
	public set name(value: string) {
		this._name = value;
	}

    /**
     * Setter description
     * @param {string} value
     */
	public set description(value: string) {
		this._description = value;
	}

    /**
     * Setter place_category
     * @param {PlaceCategory} value
     */
	public set place_category(value: PlaceCategory) {
		this._place_category = value;
	}

    /**
     * Setter place_details
     * @param {PlaceDetails} value
     */
	public set place_details(value: PlaceDetails) {
		this._place_details = value;
	}

    /**
     * Setter activities
     * @param {Activity[]} value
     */
	public set activities(value: Activity[]) {
		this._activities = value;
    }
    


    /**
     * Getter location
     * @return {LatLng}
     */
	public get location(): LatLng {
		return this._location;
	}

    /**
     * Setter location
     * @param {LatLng} value
     */
	public set location(value: LatLng) {
		this._location = value;
	}
    
    

    /**
     * Getter ratings
     * @return {Rating[]}
     */
	public get ratings(): PlaceRating[] {
		return this._ratings;
	}

    /**
     * Setter ratings
     * @param {Rating[]} value
     */
	public set ratings(value: PlaceRating[]) {
		this._ratings = value;
	}
    

    /**
     * Getter comments
     * @return {PlaceComment[]}
     */
	public get comments(): PlaceComment[] {
		return this._comments;
	}

    /**
     * Getter images
     * @return {string[]}
     */
	public get images(): string[] {
		return this._images;
	}


    /**
     * Getter images
     * @return {string[]}
     */
	public get imagesUrl(): string[] {
		return this._imagesUrl;
	}


    /**
     * Getter marker
     * @return {Marker}
     */
	public get marker(): Marker {
		return this._marker;
	}

    /**
     * Setter comments
     * @param {PlaceComment[]} value
     */
	public set comments(value: PlaceComment[]) {
		this._comments = value;
	}

    /**
     * Setter images
     * @param {string[]} value
     */
	public set images(value: string[]) {
		this._images = value;
    }
    
     /**
     * Setter images
     * @param {string[]} value
     */
	public set imagesUrl(value: string[]) {
		this._imagesUrl = value;
	}

    /**
     * Setter marker
     * @param {Marker} value
     */
	public set marker(value: Marker) {
		this._marker = value;
    }
    


    /**
     * Getter user
     * @return {User}
     */
	public get user(): UserPlace {
		return this._user;
	}

    /**
     * Setter user
     * @param {User} value
     */
	public set user(value: UserPlace) {
		this._user = value;
	}
    


    /**
     * Getter created_at
     * @return {Date}
     */
	public get created_at(): Date {
		return this._created_at;
	}

    /**
     * Setter created_at
     * @param {Date} value
     */
	public set created_at(value: Date) {
		this._created_at = value;
	}
    


    /**
     * Getter updated_at
     * @return {Date}
     */
	public get updated_at(): Date {
		return this._updated_at;
	}

    /**
     * Setter updated_at
     * @param {Date} value
     */
	public set updated_at(value: Date) {
		this._updated_at = value;
	}
    
    


    /**
     * Getter entry_registers
     * @return {EntryRegister}
     */
	public get checkin_registers(): EntryRegister[] {
		return this._checkin_registers;
	}

    /**
     * Setter entry_registers
     * @param {EntryRegister} value
     */
	public set checkin_registers(value: EntryRegister[]) {
		this._checkin_registers = value;
	}
    


    isFree(): boolean {
        return this.place_details.free;
    }

    hasWater(): boolean {
        return this.place_details.water;
    }

    hasWC(): boolean {
        return this.place_details.wc;
    }



}