import { Component } from '@angular/core';
import { IonicPage, NavController, NavParams, ModalController, LoadingController, DateTime } from 'ionic-angular';
import { Place } from '../../models/place/place';
import { LatLng} from '@ionic-native/google-maps';
import { assertNotNull } from '@angular/compiler/src/output/output_ast';
import { Photo, PlaceInterface, PlaceRating } from '../../providers/utils/constants';
import { LocationsService } from '../../providers/locations-service/locations-service';
import { GalleryModal } from 'ionic-gallery-modal';
import { AlertService } from '../../providers/alert-service/alert-service';
import { PlacesService } from '../../providers/places-service/places-service';
import { User } from '../../models/user/user';
import { AuthService } from '../../providers/auth-service/auth-service';
import { Observable } from 'rxjs/observable';
import { UserServiceProvider } from '../../providers/user-service/user-service';
import { Subscription } from 'rxjs/Subscription';

/**
 * Generated class for the DetailPlacePage page.
 *
 * See https://ionicframework.com/docs/components/#navigation for more info on
 * Ionic pages and navigation.
 */

@IonicPage()
@Component({
  selector: 'page-detail-place',
  templateUrl: 'detail-place.html',
})
export class DetailPlacePage {

  currentUser: User;

  userSubscription: Subscription;

  currentItem: Place;

  favIcon: any = {
    name: null,
    color: null,
    active: null
  }

  ratingOptions: any = {
    readonly: true
  }

  rating: any;

  distance: string;

  defaultImage = 'https://www.placecage.com/1000/1000';
  
  offset = 200;

  avatars: any[] = [];

  newComment: string;

  enableInput = false;

  constructor(
    public navCtrl: NavController, 
    public navParams: NavParams,
    public locationsService: LocationsService,
    public modalCtrl: ModalController,
    public loadingCtrl: LoadingController,
    public alert: AlertService,
    public placeService: PlacesService,
    public auth: AuthService,
    public userService: UserServiceProvider
  ) {

    this.favIcon = {
      name: 'flag',
      color: null,
      active: false
    }

    this.rating = {};
    this.distance = null;
    
  }

  ionViewDidLoad() {
    console.log('CurrentItem: ', this.currentItem);
    this.userSubscription = this.auth.subscribeUser().subscribe( user => {
      this.currentUser = user;
    })

    const loader = this.loadingCtrl.create();
    loader.present();

    let item = this.navParams.get('item');
    let placeId = item ? item.uid : 'GzOGWQPj9SbdZfYUWG25';
    this.placeService.subscription(placeId).subscribe( (data: PlaceInterface) => {
      loader.dismiss();

      this.currentItem = new Place();
      this.currentItem.setData(data);
      this.avatars = this.currentItem.checkin_registers;
    });

    
    this.distanceNearby();
  }

  getViewTitle(): string {
    return this.currentItem ? this.currentItem.name : '';
  }


  followPlace() {
    this.favIcon.active = !this.favIcon.active;

    if (this.favIcon.active) {
      this.favIcon.color = 'assertive';
    } else {
      this.favIcon.color = null;
    }
  }



  distanceNearby() {
    let place = this.currentItem;
    if (place && place.location) {
      this.locationsService.applyHaversine([place]).subscribe( (result) => {
        this.currentItem['distance'] = result[0].distance;
      })
    }
  }


  isFree(): string {
    return this.currentItem.isFree() ? 'YES' : 'NO';
  }

  hasWater(): string {
    return this.currentItem.hasWater() ? 'YES' : 'NO';
  }

  hasWC(): string {
      return this.currentItem.hasWC() ? 'YES' : 'NO';
  }

  goToDescription() {
    this.navCtrl.push('DescriptionPage', {desc: this.currentItem.description});
  }

  showImage(index) {
    let modalOptions = {
      photos: this.getPhotos(),
      initialSlide: index,
      closeIcon: 'close'
    }
    let modal = this.modalCtrl.create(GalleryModal, modalOptions);
    modal.present();
  }

  private getPhotos() {
    let result = [];
    for (let img of this.currentItem.imagesUrl) {
      let photo: Photo = {
        url: img,
        type: 'url'
      }
      result.push(photo);
    }
    return result;
  }

  avatarsElement(): any {
    this.avatars.length > 0;
    return  {
      show: this.avatars.length > 0,
      show2: this.avatars.length > 1, 
      show3: this.avatars.length > 2, 
      show4: this.avatars.length > 3, 
      show5: this.avatars.length > 4, 
      showMore: this.avatars.length > 5
    }
  }


  goToAvatarsList() {

  }


  onStayHere() {
    if( !this.auth.authenticated ){
        this.getLogin().subscribe( () => {
          setTimeout(() => {
             this.checkInHere();
          }, 500);
        })
    } else {
      this.checkInHere();
    }
  }

  getLogin(): Observable<any> {
    return new Observable( (observer) => {
        const modal = this.modalCtrl.create('LoginPage');
        modal.onDidDismiss( (data) => {
            if ( data ) {
              observer.next(data);
            }
            observer.complete();
        })
        modal.present();
    })
  }

  private checkInHere() {
    const loader = this.loadingCtrl.create();
    loader.present();
    this.placeService.checkInHere(this.currentItem).then( (place) => {
      console.log(place);
        this.userService.checkInHere(place).then( () => {
            loader.dismiss();
            this.goToRatingPlace();
        }, (error) => {
            loader.dismiss();
            this.alert.showError(error.message);
        })
    }, (error) => {
      loader.dismiss();
      this.alert.showError(error.message);
    });
  }


  goToRatingPlace() {
    const modal = this.modalCtrl.create('RatingPlacePage');
    modal.onDidDismiss( (data) => {
      if (data) {
        const loader = this.loadingCtrl.create();
        loader.present();

        this.placeService.ratingPlace(this.currentItem, data.rating, data.comment).then( () => {
          loader.dismiss();
        }, (error) => {
          loader.dismiss();
          this.alert.showError(error.message);
        })
      }
    });
    modal.present();
  }


  enableBtnHaveBeenHere(): boolean {
    if (!this.currentItem) {
        return false;
    } else if (this.currentUser) {
        let found = this.currentItem.checkin_registers.filter( (item) => item.user.uid == this.currentUser.uid)[0];
        if (found) {
          return false;
        } else {
          return true;
        }
    } else {
        return true;
    }
    
  }

  onAddComment(msg: string) {
    if( !this.auth.authenticated ){
        this.getLogin().subscribe( () => {
          setTimeout(() => {
            this.addComment(msg);
         }, 1000);
        })
    } else {
      this.addComment(msg);
    }
  }


  private addComment(msg: string) {
      this.newComment = "";

      let loader  = this.loadingCtrl.create();
      loader.present();
      this.placeService.addComment(this.currentItem, msg).then( () => {
          loader.dismiss();
      }, (error) => {
          loader.dismiss();
      })

      this.enableInput = false;
  }


  ionViewWillUnload() {
    this.userSubscription.unsubscribe();
  }
}
