import { Component } from '@angular/core';
import { NavController, IonicPage, LoadingController } from 'ionic-angular';
import { TranslateService } from '@ngx-translate/core';
import { Place } from '../../models/place/place';
import { PlacesService } from '../../providers/places-service/places-service';
import { AlertService } from '../../providers/alert-service/alert-service';
import { PlaceRating } from '../../providers/utils/constants';
import { LocationsService } from '../../providers/locations-service/locations-service';
import { Subscription } from 'rxjs/Subscription';

@IonicPage()
@Component({
  selector: 'page-home',
  templateUrl: 'home.html'
})
export class HomePage {

  viewTitle: string = 'DISCOVER';

  places: Place[];

  loading: boolean;

  distance

  placesSubscription: Subscription;


  constructor(
    public navCtrl: NavController, 
    public placeService: PlacesService,
    public loadingCtrl: LoadingController,
    public alert: AlertService,
    public locationsService: LocationsService
  ) {

    this.loading = true;
  }

  ionViewDidLoad() {
    const loader = this.loadingCtrl.create();
    loader.present();
    this.placeService.getPlaces();
    this.placesSubscription = this.placeService.subscriptionPlaces().subscribe( (places) => {
        if (places) {
            if(places.length > 0) {
                  loader.dismiss();
                  this.loading = false;
                  this.locationsService.applyHaversine(places).subscribe( (result) => {
                    this.places = result;
                    this.places.sort( (locationA, locationB) => {
                      return locationA['distance'] - locationB['distance'];
                    })
                  })

            } else {
              loader.dismiss();
              this.loading = false;
              this.places = places;
            }
        }
    }, (error) => {
      loader.dismiss();
      this.loading = false;
      this.alert.showError(error.message);
    })
  }

  doRefresh(refresher) {
    this.placeService.getPlaces().then( () => {
      refresher.complete();
    }).catch( (error) => {
      refresher.complete();
      this.alert.showError(error.message);
    });
  }


  showEmptyResults() {
    return this.loading == false && this.places && this.places.length == 0;
  }



  distanceNearby(currentItem) {
    let place = currentItem;
    if (place && place.location) {
      this.locationsService.applyHaversine([place]).subscribe( (result) => {
        currentItem['distance'] = result[0].distance;
      })
    }
  }

  goToDetail($event) {
    let item = $event.item;
    this.navCtrl.push('DetailPlacePage', {item: item});
  }

}
