import { Component } from '@angular/core';
import { IonicPage, NavController, NavParams, ViewController, LoadingController, Loading } from 'ionic-angular';
import { FormGroup, FormBuilder, FormControl, Validators } from '@angular/forms';
import { AuthService } from '../../providers/auth-service/auth-service';
import { AlertService } from '../../providers/alert-service/alert-service';


/**
 * Generated class for the LoginPage page.
 *
 * See https://ionicframework.com/docs/components/#navigation for more info on
 * Ionic pages and navigation.
 */

@IonicPage()
@Component({
  selector: 'page-login',
  templateUrl: 'login.html',
})
export class LoginPage {

  public validations_form: FormGroup;

  private loader: Loading;

  public submitAttempt: boolean = false;

  public validation_messages = {
    'email': [
      { type: 'required', message: 'VALIDATOR.EMAIL_REQUIRED' }
    ],
    'password': [
      { type: 'required', message: 'VALIDATOR.PASSWORD_REQUIRED' },
      { type: 'minlength', message: 'VALIDATOR.PASSWORD_MINLENGTH' }
    ]
  };

  constructor(
    public navCtrl: NavController, 
    public navParams: NavParams,
    public viewCtrl: ViewController,
    public formBuilder: FormBuilder,
    public loadingCtrl: LoadingController,
    public auth: AuthService,
    public alert: AlertService
  ) {
  }

  ionViewWillLoad() {
    this.validations_form = this.formBuilder.group({
      email: new FormControl('', Validators.compose([
        Validators.required
      ])),
      password: new FormControl('', Validators.compose([
        Validators.required,
        Validators.minLength(6)
      ]))
    })
  }

  ionViewDidLoad() {
    console.log('ionViewDidLoad LoginPage');
  }


  onSubmit(values) {
    this.submitAttempt = true;

    this.loader = this.loadingCtrl.create();
    this.loader.present();
    this.auth.login(values.email, values.password)
              .then(this.responseLogin, this.errorLogin);
  }

  loginWithGoogle() {
    this.loader = this.loadingCtrl.create();
    this.loader.present();
    this.auth.signInWithGoogle()
      .then(this.responseLogin, this.errorLogin);
  }

  private responseLogin = ( isLogged ) => {
    this.loader.dismiss();

    if( isLogged ) {
      this.dismiss(true);
    }
  }

  private errorLogin = ( error ) => {
    this.loader.dismiss();

    this.alert.showError(error.message);
  }


  goToSignUp() {
    this.navCtrl.push('SignupPage', {callback: this.callbackSingup});
  }


  private callbackSingup = (isLogged) => {
    if (isLogged) {
      this.dismiss(true);
    }
  }


  dismiss(data?: any) {
    this.viewCtrl.dismiss(data);
  }
}
