import { Component, ViewChild, ElementRef } from '@angular/core';
import { IonicPage, NavController, NavParams, ViewController, Platform, ModalController, LoadingController, ActionSheetController, Content } from 'ionic-angular';
import { GoogleMapsService } from '../../providers/google-maps/google-maps';
import { LatLng, GoogleMapOptions, GoogleMaps, GoogleMapsEvent } from '@ionic-native/google-maps';
import { Observable } from 'rxjs/Observable';
import { AngularFirestore, AngularFirestoreCollection } from 'angularfire2/firestore';
import { PlacesService } from '../../providers/places-service/places-service';
import { PlaceInterface } from '../../providers/utils/constants';
import { forkJoin } from 'rxjs/Observable/forkJoin';
import { TranslateService } from '@ngx-translate/core';
import { CameraService } from '../../providers/camera-service/camera-service';

/**
 * Generated class for the NewPlacePage page.
 *
 * See https://ionicframework.com/docs/components/#navigation for more info on
 * Ionic pages and navigation.
 */

enum Types {
    PLACE_CATEGORIES = 'place_categories',
    PLACE_DETAILS  = 'place_details',
    ACTIVITIES = 'activities'
}

interface Place {
  currentPosition: LatLng;
  name: string;
  description: string;
  place_category: any;
  place_details: any;
  activities: any[];
}

interface PlaceCategory {
  name: string;
}
@IonicPage()
@Component({
  selector: 'page-new-place',
  templateUrl: 'new-place.html',
})
export class NewPlacePage {
  @ViewChild(Content) content: Content;

  isDevice: boolean;

  data: PlaceInterface = {
    uid: null,
    location: null,
    name: null,
    description: null,
    place_category: null,
    place_details: {
      free: false,
      water: false,
      wc: false
    },
    activities: null,
    images: null,
    user: null,
    created_at: null,
    updated_at: null,
    checkin_registers: null,
    ratings: null,
    comments: null,
    marker: null
  }

  place_categories: any = [];
  activities: any[] = [];

  constructor(
    public navCtrl: NavController, 
    public navParams: NavParams,
    public viewCtrl: ViewController,
    public maps: GoogleMapsService,
    public platform: Platform,
    public modalCtrl: ModalController,
    public placeService: PlacesService,
    public loadingCtrl: LoadingController,
    public actionSheetCtrl: ActionSheetController,
    public translate: TranslateService,
    public cameraService: CameraService
  ) {

    this.isDevice = this.platform.is('cordova');

    this.data.images = [];
    let currentPlace = this.navParams.get('place');
    if (currentPlace) {
      this.data =  {
        uid: currentPlace.uid,
        location: currentPlace.location,
        name: currentPlace.name,
        description: currentPlace.description,
        place_category: currentPlace.place_category,
        place_details: currentPlace.place_details,
        activities: currentPlace.activities,
        user: currentPlace.user,
        images: currentPlace.images,
        created_at: currentPlace.created_at,
        ratings: currentPlace.ratings,
        comments: currentPlace.comments,
        marker: currentPlace.marker,
        updated_at: currentPlace.updated_at,
        checkin_registers: currentPlace.checkin_registers
      }
    }
  }

  ionViewDidLoad() {
    if ( this.isDevice ) {
      this.data.location = this.maps.pointMarker.getPosition() as LatLng;
    }

    let categoriesLoaded = this.placeService.getPlaceCategories();
    let activitiesLoaded = this.placeService.getActivities();

    const loader = this.loadingCtrl.create();
    loader.present();
    categoriesLoaded.subscribe( (data) => {
      this.place_categories = data;
      activitiesLoaded.subscribe( (data) => {
        this.activities = data;
        loader.dismiss();
      })
    })
  
  }

  addPhotos() {
    this.presentActionSheet();
  }


  save() {
    console.log(Object.assign({},this.data));
    // this.navCtrl.push('BlankPage', {data: Object.assign({},this.data)});
    const loader = this.loadingCtrl.create();
    loader.present();

    this.placeService.save(this.data).then( (place) => {
      loader.dismiss();
       this.dismiss(place);
    }, (error) => {
      loader.dismiss();
      console.error(error);
    })
  }

  dismiss(data?){
    this.viewCtrl.dismiss(data);
  }


  deleteActivity(index) {
    this.data.activities.splice(index, 1);
  }

  deletePhoto(index) {
    this.data.images.splice(index, 1);
  }



  scrollToBottom() {
    setTimeout( () => {
      this.content.scrollToBottom();
    }, 500);
  }


  presentActionSheet() {
    let cameraBtn = '';
    let libraryBtn = '';
    this.translate.get( 'BUTTONS').subscribe( (value) => {
      cameraBtn = value.CAMERA;
      libraryBtn = value.LIBRARY;
    });
    let actionSheet = this.actionSheetCtrl.create({
      buttons: [
        {
          text: cameraBtn,
          handler: () => {
            const loader = this.loadingCtrl.create();
            loader.present();
            this.cameraService.getPictureFromCamera().then( (base64) => {
              console.log( base64);
              this.data.images.push(base64);
              this.scrollToBottom();
              loader.dismiss();
            })
          }
        },
        {
          text: libraryBtn,
          handler: () => {
            const loader = this.loadingCtrl.create();
            loader.present();
            this.cameraService.getPictureFromPhotoLibrary().then( (base64) => {
              console.log( base64);
              this.data.images.push(base64);
              this.scrollToBottom();
              loader.dismiss();
            })
          }
        }
      ]
    });
 
    actionSheet.present();
  }

}
