import { Component } from '@angular/core';
import { IonicPage, NavController, NavParams, Loading, LoadingController } from 'ionic-angular';
import { FormGroup, FormBuilder, Validators, FormControl } from '@angular/forms';
import { AuthService } from '../../providers/auth-service/auth-service';
import { AlertService } from '../../providers/alert-service/alert-service';

/**
 * Generated class for the SignupPage page.
 *
 * See https://ionicframework.com/docs/components/#navigation for more info on
 * Ionic pages and navigation.
 */

@IonicPage()
@Component({
  selector: 'page-signup',
  templateUrl: 'signup.html',
})
export class SignupPage {

  public validations_form: FormGroup;

  private loader: Loading;

  public submitAttempt: boolean = false;

  public callbackToLogin: any;

  public validation_messages = {
    'username': [
      { type: 'required', message: 'VALIDATOR.USERNAME_REQUIRED' }
    ],
    'email': [
      { type: 'required', message: 'VALIDATOR.EMAIL_REQUIRED' }
    ],
    'password': [
      { type: 'required', message: 'VALIDATOR.PASSWORD_REQUIRED' },
      { type: 'minlength', message: 'VALIDATOR.PASSWORD_MINLENGTH' }
    ],
    'passwordConfirm': [
      { type: 'required', message: 'VALIDATOR.CONFIRM_REQUIRED' },
      { type: 'equalValue', message: 'VALIDATOR.EQUALS_PASSWORD'}
    ]
  };

  constructor(
    public navCtrl: NavController, 
    public navParams: NavParams,
    public formBuilder: FormBuilder,
    public loadingCtrl: LoadingController,
    public auth: AuthService,
    public alert: AlertService
  ) {

    this.callbackToLogin = this.navParams.get('callback');
  }

  ionViewWillLoad() {
    this.validations_form = this.formBuilder.group({
      username: new FormControl('', Validators.compose([
        Validators.required
      ])),
      email: new FormControl('', Validators.compose([
        Validators.required
      ])),
      password: new FormControl('', Validators.compose([
        Validators.required,
        Validators.minLength(6)
      ])),
      passwordConfirm: new FormControl('', Validators.compose([
        Validators.required
      ]))

    });
  }


  ionViewDidLoad() {
    console.log('ionViewDidLoad SignupPage');
  }

  onSubmit(values) {
    this.submitAttempt = true;
    if ( values.password !== values.passwordConfirm ) {
      this.validations_form.get('passwordConfirm').setErrors( {"equalValue": true} );
    }


    if ( this.validations_form.valid ) {
        this.loader = this.loadingCtrl.create();
        this.loader.present();
    
        this.auth.signup(values.username, values.email, values.password)
                  .then(this.responseSignup, this.errorSignup);
    }

  }

  private responseSignup = ( isLogged ) => {
    this.loader.dismiss();

    if( isLogged ) {
      this.callbackToLogin(true);
      this.navCtrl.pop();
    }
  }

  private errorSignup = ( error ) => {
    this.loader.dismiss();

    console.error(error);
    this.alert.showError(error.message);
  }

}
