import { HttpClient } from '@angular/common/http';
import { Injectable } from '@angular/core';
import { Camera, CameraOptions } from '@ionic-native/camera';

/*
  Generated class for the CameraServiceProvider provider.

  See https://angular.io/guide/dependency-injection for more info on providers
  and Angular DI.
*/
@Injectable()
export class CameraService {

  constructor(public http: HttpClient, private camera: Camera) {
    console.log('Hello CameraService Provider');
  }

  getPictureFromCamera(crop?: boolean) {
    return this.getImage(this.camera.PictureSourceType.CAMERA);
  }

  getPictureFromPhotoLibrary(crop?: boolean) {
    return this.getImage(this.camera.PictureSourceType.PHOTOLIBRARY);
  }
  

  private getImage(pictureSourceType): Promise<any> {

    const options : CameraOptions = {
      quality: 50, // picture quality
      destinationType: this.camera.DestinationType.DATA_URL,
      encodingType: this.camera.EncodingType.JPEG,
      mediaType: this.camera.MediaType.PICTURE,
      sourceType: pictureSourceType,
      saveToPhotoAlbum: pictureSourceType == this.camera.PictureSourceType.PHOTOLIBRARY ? false : true,
      correctOrientation: true
    }
    
    return new Promise( (resolve, reject) => {
      this.camera.getPicture(options).then((imageData) => {
        let base64Image = 'data:image/jpg;base64,' + imageData;
        resolve(base64Image);
       }, (err) => {
        reject(err);
       });
    })
    
  }

}
