import { HttpClient } from '@angular/common/http';
import { Injectable } from '@angular/core';
import { Platform } from 'ionic-angular';
import { Network } from '@ionic-native/network';
import { Subscription } from 'rxjs/Subscription';
import { BehaviorSubject } from 'rxjs/BehaviorSubject';
import { Observable } from 'rxjs/Observable';

declare var Connection;

/*
  Generated class for the ConnectivityProvider provider.

  See https://angular.io/guide/dependency-injection for more info on providers
  and Angular DI.
*/
@Injectable()
export class Connectivity {

  onDevice: boolean;

  online: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(true);
  online$: Observable<boolean> = this.online.asObservable();

  connectSubscription: Subscription;
  disconnectSubscription: Subscription;
 
  constructor(public platform: Platform, public network: Network){
    this.onDevice = this.platform.is('cordova');

    this.addListeners();
  }

  private addListeners() {
    this.connectSubscription = this.network.onConnect().subscribe( () => {
      this.online.next(true);
    });

    this.disconnectSubscription = this.network.onDisconnect().subscribe( () => {
      this.online.next(false);
    })
  }
 
  isOnline(): boolean {
    return this.online.getValue();
  }
 
  isOffline(): boolean {
    return !this.online.getValue();
  }

  disableListeners() {
    this.connectSubscription.unsubscribe();
    this.disconnectSubscription.unsubscribe();
  }

}
