import { HttpClient } from '@angular/common/http';
import { Injectable, NgZone } from '@angular/core';
import { Connectivity } from '../connectivity/connectivity';
import { Observable } from 'rxjs/Observable';
import { GoogleMaps, GoogleMap, GoogleMapsEvent, GoogleMapOptions, GoogleMapsAnimation, LatLng, MyLocation, Marker } from '@ionic-native/google-maps';
import { BehaviorSubject } from 'rxjs/BehaviorSubject';
import { TranslateService } from '@ngx-translate/core';
import { Place } from '../../models/place/place';

declare var google;
/*
  Generated class for the GoogleMapsProvider provider.

  See https://angular.io/guide/dependency-injection for more info on providers
  and Angular DI.
*/
@Injectable()
export class GoogleMapsService {

  private mapElement: any;
  private pleaseConnect: any;
  private searchBar: any;
  map: GoogleMap;
  mapInitialised: boolean = false;
  pointMarker: Marker;
  markers: any[] = [];
  private _activeMarker: Marker;

  cities: any[];
  city: any;
  autocompleteService: any;
  placesService: any;
  searchDisabled: boolean;

  private clickPosition: BehaviorSubject<any> = new BehaviorSubject<any>(null);
  clickPosition$ = this.clickPosition.asObservable();

  apiKey: string = 'AIzaSyDkSIj0syJxCda6bc243RjRkkrgMBX65NM';


  constructor(public connectivity: Connectivity, public zone: NgZone, public translate: TranslateService) {
    console.log('Hello GoogleMapsProvider Provider');
  }

  init(mapElement: any, pleaseConnect: any, searchBar: any): Observable<any> {

    this.mapElement = mapElement;
    this.pleaseConnect = pleaseConnect;
    this.searchBar = searchBar;

    return this.loadGoogleMaps();

  }

  setPleaseConnectDiv(pleaseConnect: any) {
    this.pleaseConnect = pleaseConnect;
  }



  loadGoogleMaps(): Observable<any> {
     
    return new Observable( (observer) => {
      if(typeof google == "undefined" || typeof google.maps == "undefined"){
 
        console.log("Google maps JavaScript needs to be loaded.");
        this.disableMap();
 
        if(this.connectivity.isOnline()){
 
          window['mapInit'] = () => {
 
            this.initMap().subscribe(() => {
              observer.next(true);
              observer.complete();
            });
 
            this.enableMap();
          }
 
          let script = document.createElement("script");
          script.id = "googleMaps";
 
          if(this.apiKey){
            script.src = 'http://maps.google.com/maps/api/js?key=' + this.apiKey + '&libraries=places';
          } else {
            script.src = 'http://maps.google.com/maps/api/js?libraries=places';      
          }
 
          document.body.appendChild(script); 
 
        }
      } else {

        if(this.connectivity.isOnline()) {
          this.initMap().subscribe(() => {
            observer.next(true);
            observer.complete();
          })
  
          this.enableMap();
        } else {
          this.disableMap();
        }
    
      
      }

     
      this.addConnectivityListeners();
      
    })

  }


  initMap(): Observable<any> {
    this.mapInitialised = true;

    return new Observable( (observer) => {

      let mapOptions = this.getMapOptions();
      this.map = GoogleMaps.create(this.mapElement, mapOptions);

      this.map.one(GoogleMapsEvent.MAP_READY).then( () => {

        this.addGoogleMapsListeners();

        this.initAutocompleteService();

        observer.next(true);
        observer.complete();
      })
    })
  }

  getMapOptions() {
    let mapOptions: GoogleMapOptions = {
      backgroundColor: 'white',
      controls: {
        compass: true,
        myLocationButton: true,
        myLocation: true,
        indoorPicker: true,
        mapToolbar: false,
        zoom: true
      },
      gestures: {
        scroll: true,
        tilt: true,
        rotate: true,
        zoom: true
      },
      camera: {
        target: {
          lat: 40.483102,
          lng: -1.3335809
        },
        zoom: 5,
        tilt: 0
      }
    }
    return mapOptions;
  }


  getPosition() {
    this.getMyLocation().then( (response: MyLocation) => {
      this.map.set('myPosition', response.latLng);
      this.setCenter(response.latLng, 13);
    })   
  }

  getMyLocation() {
    return new Promise((resolve, reject) => {
      this.map.getMyLocation().then( (response: MyLocation) => {
        resolve(response);
      }).catch( (error) => {
        reject(error);
      })
    })
  }

  set activeMarker(marker: Marker) {
    this._activeMarker = marker;
  }

  get activeMarker(): Marker {
    return this._activeMarker;
  }

  closeTitleActiveMarker() {
    if( this._activeMarker ) {
      this._activeMarker.hideInfoWindow();
    }
  }

  removeActiveMarker() {
    this.closeTitleActiveMarker();
    this._activeMarker = null;
  }

  removePointMarker() {
    if(this.pointMarker) {
      this.pointMarker.remove();
    }
  }

  addGoogleMapsListeners() {
    this.map.on( GoogleMapsEvent.MAP_CLICK ).subscribe( (latLng) => {
      this.clickPosition.next(latLng[0]);
      this.removePointMarker();
    });
    

    this.map.on('camera_target_changed').subscribe( (latLng) => {
      console.log("map center = " + latLng);
    });

   
    this.map.on('myPosition_changed').subscribe( (latLng) => {
      console.log('MyPosition_changed: ', latLng);
    })
  }

 

  addMarkerPoint(_latLng?) {
    console.log('Draw a circle');
    let title = '';
    this.translate.get('MAP_PAGE.UNKNOWN_PLACE').subscribe( (value)=>{
      title = value;
    })

    let latLng = _latLng && _latLng[0] ? _latLng[0] : new LatLng(42.0660285,2.253194); 
    
    this.removePointMarker();
    
    var image = {
      url: 'assets/imgs/point.png',
      // This marker is 20 pixels wide by 32 pixels high.
      size: new google.maps.Size(32, 32),
      // The origin for this image is (0, 0).
      origin: new google.maps.Point(0, 0),
      // The anchor for this image is the base of the flagpole at (0, 32).
      anchor: new google.maps.Point(16, 16)
    };

    return new Promise( (resolve, reject) => {
      this.map.addMarker({
        title: title,
        position: latLng,
        icon: image,
        index: -1,
        newPlace: true
      }).then( (marker) => {
        console.log(marker);
        this.pointMarker = marker;
        this.setCenter(latLng);
        resolve(marker);
      })
    })
    
  }


  disableMap(): void {
 
    if(this.pleaseConnect){
      this.pleaseConnect.style.display = "block";
    }
 
  }
 
  enableMap(): void {
 
    if(this.pleaseConnect){
      this.pleaseConnect.style.display = "none";
    }
 
  }

  addConnectivityListeners(): void {

    this.connectivity.online$.subscribe( (isOnline) => {

      if (isOnline) {

        setTimeout( () => {

          if(!this.mapInitialised) {
            this.initMap().subscribe();
          }

          this.enableMap();
        }, 1000);

      } else {

        console.log("offline");

        this.disableMap();

      }
    })
  }


  clearMarkers() {
    this.markers = [];
    this.map.clear();
  }


  addMarker(lat: number, lng: number, title: string, index: string): Promise<any>  {
    let latLng = new LatLng(lat, lng);
    var image = {
      url: 'assets/imgs/markerIcon.png',
      // This marker is 20 pixels wide by 32 pixels high.
      size: new google.maps.Size(13, 32),
      // The origin for this image is (0, 0).
      origin: new google.maps.Point(0, 0),
      // The anchor for this image is the base of the flagpole at (0, 32).
      anchor: new google.maps.Point(6, 32)
    };
    return new Promise( (resolve) => {
      this.map.addMarker({
        title: title,
        icon: image,
        animation: GoogleMapsAnimation.BOUNCE,
        position: latLng,
        index: index
      }).then( (marker) => {
        this.markers.push(marker);
        resolve(marker);
      });
    });
  }

  getMarker(index) {
    return this.markers.filter( (marker) => marker.get('index') === index )[0];
  }

  setCenter(latLng: LatLng, zoom?: number) {
    // this.map.setCameraTarget( latLng );
    // this.map.setCameraZoom(zoom | 13);
    this.map.animateCamera({
      target: latLng,
      zoom: zoom,
      duration: 1000
    })
  }

  initAutocompleteService() {
    this.autocompleteService = new google.maps.places.AutocompleteService();

    this.placesService = new google.maps.places.PlacesService(this.mapElement);
    this.searchDisabled = false;
  }


  selectCity(place) {

    this.searchBar.blur();

    this.cities = [];

    let location = {
      lat: null,
      lng: null,
      name: place.nambe
    };

    this.placesService.getDetails( {placeId: place.place_id}, (details) => {

      this.zone.run( () => {

        location.name = details.name;
        location.lat = details.geometry.location.lat();
        location.lng = details.geometry.location.lng();

        this.setCenter(new LatLng(location.lat, location.lng), 13);

        this.city = location;

      })
    })
  }


  searchCity(query, callback) {

    if (query.length > 0 && !this.searchDisabled) {

      let config = {
        types: ['(cities)'],
        input: query
      }

      this.autocompleteService.getPlacePredictions(config, (predictions, status) => {

        if (status == google.maps.places.PlacesServiceStatus.OK && predictions) {

          this.cities = [];

          predictions.forEach(prediction => {
            this.cities.push(prediction);
          });
          callback(this.cities);
        }
      });


    } else {
      this.cities = [];
      callback(this.cities);
    }
  }

}
