import { HttpClient } from '@angular/common/http';
import { Injectable } from '@angular/core';
import { Observable } from 'rxjs/Observable';
import 'rxjs/add/observable/of';
import 'rxjs/add/operator/map';
import { GoogleMapsService } from '../google-maps/google-maps';
import { Geolocation } from '@ionic-native/geolocation';
import { LatLng } from '@ionic-native/google-maps';

/*
  Generated class for the LocationsServiceProvider provider.

  See https://angular.io/guide/dependency-injection for more info on providers
  and Angular DI.
*/
@Injectable()
export class LocationsService {

  data: any;

  constructor(public http: HttpClient, public maps: GoogleMapsService, public geolocation: Geolocation) {
    console.log('Hello LocationsServiceProvider Provider');
  }

  load() {

    if(this.data) {
      return Observable.of(this.data);
    } 

    return new Observable( (observer) => {
      this.http.get('assets/data/locations.json').subscribe( (data: any) => {
        console.log(data);
        this.applyHaversine(data.locations).subscribe( (result) => {
          this.data = result;
          this.data.sort( (locationA, locationB) => {
            return locationA.distance - locationB.distance;
          })
          observer.next(this.data);
          observer.complete();
        })
      })
    })
  }


  applyHaversine(locations): Observable<any>{

    return new Observable( (observer) => {
      setTimeout(() => {
        this.geolocation.getCurrentPosition().then( (response: any) => {
          console.log(response);

          let usersLocation = new LatLng(response.coords.latitude, response.coords.longitude);
  
          locations.map( (place) => {
  
            let placeLocation = {
              lat: place.location.lat,
              lng: place.location.lng
            };
      
            place.distance = this.getDistanceBetweenPoints(
              usersLocation,
              placeLocation,
              'km'
            ).toFixed(2);
          });
  
          observer.next(locations);
        }).catch( (error) => {
          console.log(error);
        })
      }, 1000);
      
    })
  }


  getDistanceBetweenPoints( start, end, units) {
    let earthRadius = {
      miles: 3958.8,
      km: 6371
    };

    let R = earthRadius[ units || 'km' ];
    let lat1 = start.lat;
    let lon1 = start.lng;
    let lat2 = end.lat;
    let lon2 = end.lng;

    let dLat = this.toRad((lat2 - lat1));
    let dLon = this.toRad((lon2 - lon1));
    let a = Math.sin(dLat / 2) * Math.sin(dLat / 2) +
    Math.cos(this.toRad(lat1)) * Math.cos(this.toRad(lat2)) *
    Math.sin(dLon / 2) *
    Math.sin(dLon / 2);
    let c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a));
    let d = R * c;

    return d;
  }

  toRad(x) {
    return x * Math.PI / 180;
  }


}
