import { HttpClient } from '@angular/common/http';
import { Injectable } from '@angular/core';
import { AuthService } from '../auth-service/auth-service';
import { CheckInPlace } from '../utils/constants';
import { Place } from '../../models/place/place';
import { User } from '../../models/user/user';
import { AngularFirestore } from 'angularfire2/firestore';
import { Subscription } from 'rxjs/Subscription';

/*
  Generated class for the UserServiceProvider provider.

  See https://angular.io/guide/dependency-injection for more info on providers
  and Angular DI.
*/
@Injectable()
export class UserServiceProvider {

  private currentUser: User;

  private userSubscription: Subscription;

  constructor(
    public http: HttpClient,
    public auth: AuthService,
    public afs: AngularFirestore,
  ) {
    console.log('Hello UserServiceProvider Provider');

    this.userSubscription = this.auth.subscribeUser().subscribe( (user) => {
        if(user) {
          this.currentUser = new User();
          this.currentUser.set(user);
        } else {
          this.currentUser = null;
        }
    })
  }


  checkInHere(_place: Place): Promise<any> {

      let place = new Place();
      place.setData(_place);

      let newObject: CheckInPlace = {
        created_at: new Date(),
        place: {
            uid: place.uid,
            name: place.name,
            location: place.location,
            image: place.images[0],
            imageUrl: place.imagesUrl[0],
            place_category: place.place_category,
            rating: place.getRating()
        }
      }

      if( !this.currentUser.checkins ) {
        this.currentUser.checkins = [];
      }
      this.currentUser.checkins.push(newObject);

      var userRef = this.afs.collection('users').doc(this.currentUser.uid);

      return userRef.set({checkins: this.currentUser.checkins}, { merge: true });
  }

}
